# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'c:\Users\macad\Documents\Git\BCML-2\.vscode\package.ui',
# licensing of 'c:\Users\macad\Documents\Git\BCML-2\.vscode\package.ui' applies.
#
# Created: Fri Sep 13 20:51:27 2019
#      by: pyside2-uic  running on PySide2 5.13.0
#
# WARNING! All changes made in this file will be lost!

from PySide2 import QtCore, QtGui, QtWidgets

class Ui_PackageDialog(object):
    def setupUi(self, PackageDialog):
        PackageDialog.setObjectName("PackageDialog")
        PackageDialog.setWindowModality(QtCore.Qt.WindowModal)
        PackageDialog.resize(320, 301)
        PackageDialog.setModal(True)
        self.verticalLayout = QtWidgets.QVBoxLayout(PackageDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.label = QtWidgets.QLabel(PackageDialog)
        self.label.setObjectName("label")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label)
        self.txtName = QtWidgets.QLineEdit(PackageDialog)
        self.txtName.setObjectName("txtName")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.txtName)
        self.label1 = QtWidgets.QLabel(PackageDialog)
        self.label1.setObjectName("label1")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label1)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.txtFolder = QtWidgets.QLineEdit(PackageDialog)
        self.txtFolder.setObjectName("txtFolder")
        self.horizontalLayout_3.addWidget(self.txtFolder)
        self.btnBrowseContent = QtWidgets.QPushButton(PackageDialog)
        self.btnBrowseContent.setStyleSheet("padding: 3px 6px;")
        self.btnBrowseContent.setObjectName("btnBrowseContent")
        self.horizontalLayout_3.addWidget(self.btnBrowseContent)
        self.formLayout.setLayout(2, QtWidgets.QFormLayout.FieldRole, self.horizontalLayout_3)
        self.label_2 = QtWidgets.QLabel(PackageDialog)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.txtImage = QtWidgets.QLineEdit(PackageDialog)
        self.txtImage.setObjectName("txtImage")
        self.horizontalLayout_2.addWidget(self.txtImage)
        self.btnBrowseImg = QtWidgets.QPushButton(PackageDialog)
        self.btnBrowseImg.setStyleSheet("padding: 3px 6px;")
        self.btnBrowseImg.setObjectName("btnBrowseImg")
        self.horizontalLayout_2.addWidget(self.btnBrowseImg)
        self.formLayout.setLayout(3, QtWidgets.QFormLayout.FieldRole, self.horizontalLayout_2)
        self.label_3 = QtWidgets.QLabel(PackageDialog)
        self.label_3.setObjectName("label_3")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.label_3)
        self.txtUrl = QtWidgets.QLineEdit(PackageDialog)
        self.txtUrl.setObjectName("txtUrl")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.txtUrl)
        self.label_4 = QtWidgets.QLabel(PackageDialog)
        self.label_4.setObjectName("label_4")
        self.formLayout.setWidget(5, QtWidgets.QFormLayout.LabelRole, self.label_4)
        self.txtDescript = QtWidgets.QPlainTextEdit(PackageDialog)
        self.txtDescript.setObjectName("txtDescript")
        self.formLayout.setWidget(6, QtWidgets.QFormLayout.SpanningRole, self.txtDescript)
        self.verticalLayout.addLayout(self.formLayout)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.btnAdvanced = QtWidgets.QPushButton(PackageDialog)
        self.btnAdvanced.setObjectName("btnAdvanced")
        self.horizontalLayout.addWidget(self.btnAdvanced)
        self.buttonBox = QtWidgets.QDialogButtonBox(PackageDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.horizontalLayout.addWidget(self.buttonBox)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(PackageDialog)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL("accepted()"), PackageDialog.accept)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL("rejected()"), PackageDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(PackageDialog)

    def retranslateUi(self, PackageDialog):
        PackageDialog.setWindowTitle(QtWidgets.QApplication.translate("PackageDialog", "Create Nano Patch Mod Package", None, -1))
        self.label.setText(QtWidgets.QApplication.translate("PackageDialog", "Name:", None, -1))
        self.label1.setText(QtWidgets.QApplication.translate("PackageDialog", "Mod folder:", None, -1))
        self.btnBrowseContent.setText(QtWidgets.QApplication.translate("PackageDialog", "Browse...", None, -1))
        self.label_2.setText(QtWidgets.QApplication.translate("PackageDialog", "Image (optional):", None, -1))
        self.btnBrowseImg.setText(QtWidgets.QApplication.translate("PackageDialog", "Browse...", None, -1))
        self.label_3.setText(QtWidgets.QApplication.translate("PackageDialog", "URL (optional):", None, -1))
        self.label_4.setText(QtWidgets.QApplication.translate("PackageDialog", "Description:", None, -1))
        self.btnAdvanced.setText(QtWidgets.QApplication.translate("PackageDialog", "Advanced...", None, -1))

